/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: imports.c,v 1.6 2006/06/13 20:36:15 phargrov Exp $
 */

#ifndef EXPORT_SYMTAB
  #define EXPORT_SYMTAB
#endif

#include "blcr_config.h"
#include "blcr_ksyms.h"

#include <linux/config.h>
#if defined(CONFIG_SMP) && ! defined(__SMP__)
  #define __SMP__
#endif
#if defined(CONFIG_MODVERSIONS) && ! defined(MODVERSIONS)
  #define MODVERSIONS
#endif
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/version.h>
#include <linux/list.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/mount.h>
#include <linux/slab.h>
#include <linux/highmem.h>
#include <linux/spinlock.h>
#include <asm/atomic.h>
#include <linux/unistd.h>
#include <asm/unistd.h>
#if HAVE_LINUX_NAMEI_H
  #include <linux/namei.h>
#endif
#if HAVE_LINUX_SYSCALLS_H
  #include <linux/syscalls.h>           /* for sys_close, etc. */
#endif
#if HAVE_LINUX_SYSCALL_H
  #include <linux/syscall.h>           /* for sys_close, etc. */
#endif


/* This comment contains "EXPORT_SYMBOL" to trigger proper kbuild magic */

#if CR_KDATA_sys_call_table
  extern void * sys_call_table[];
  CR_IMPORT_KDATA(sys_call_table)
#endif
#if defined(CR_KDATA_cpu_gdt_table) && CR_KDATA_cpu_gdt_table
  #include <asm/desc.h>
  CR_IMPORT_KDATA(cpu_gdt_table)            // i386 and x86_64
#endif
#if defined(CR_KDATA_mmu_cr4_features) && CR_KDATA_mmu_cr4_features
  #include <asm/processor.h>
  CR_IMPORT_KDATA(mmu_cr4_features)            // i386 and x86_64
#endif
#if CR_KDATA_write_pipe_fops
  extern struct file_operations write_pipe_fops;
  CR_IMPORT_KDATA(write_pipe_fops)
#endif
#if CR_KDATA_rdwr_pipe_fops
  extern struct file_operations rdwr_pipe_fops;
  CR_IMPORT_KDATA(rdwr_pipe_fops)
#endif
#if CR_KDATA_read_pipe_fops
  extern struct file_operations read_pipe_fops;
  CR_IMPORT_KDATA(read_pipe_fops)
#endif
#if CR_KDATA_write_fifo_fops
  extern struct file_operations write_fifo_fops;
  CR_IMPORT_KDATA(write_fifo_fops)
#endif
#if CR_KDATA_rdwr_fifo_fops
  extern struct file_operations rdwr_fifo_fops;
  CR_IMPORT_KDATA(rdwr_fifo_fops)
#endif
#if CR_KDATA_read_fifo_fops
  extern struct file_operations read_fifo_fops;
  CR_IMPORT_KDATA(read_fifo_fops)
#endif
#if CR_KDATA_pidmap_array
  typedef struct pidmap {
    atomic_t nr_free;
    void *page;
  } pidmap_t;
  extern pidmap_t pidmap_array[];
  CR_IMPORT_KDATA(pidmap_array)
#endif
#if defined(CR_KDATA_pidmap_lock) && CR_KDATA_pidmap_lock
  extern spinlock_t pidmap_lock;
  CR_IMPORT_KDATA(pidmap_lock)
#endif
#if CR_KDATA_anon_pipe_buf_ops
  extern struct pipe_buf_operations anon_pipe_buf_ops;
  CR_IMPORT_KDATA(anon_pipe_buf_ops)
#endif


#if CR_KCODE_sys_dup2 
  #if !HAVE_SYS_DUP2_PROTOTYPE
    /* Some kernels lack a prototype for sys_dup2() */
    extern asmlinkage long sys_dup2(unsigned int oldfd, unsigned int newfd);
  #endif
  CR_IMPORT_KCODE(sys_dup2)
#endif
#if CR_KCODE_sys_lseek 
  #if !HAVE_SYS_LSEEK_PROTOTYPE
    /* Some kernels lack a prototype for sys_lseek() */
    extern asmlinkage off_t sys_lseek(unsigned int fd, off_t offset, unsigned int origin);
  #endif
  CR_IMPORT_KCODE(sys_lseek)
#endif
#if CR_KCODE_sys_ftruncate 
  #if !HAVE_SYS_FTRUNCATE_PROTOTYPE
    /* Some kernels lack a prototype for sys_ftruncate() */
    extern asmlinkage long sys_ftruncate(unsigned int fd, unsigned long length);
  #endif
  CR_IMPORT_KCODE(sys_ftruncate)
#endif
#if CR_KCODE_sys_close
  CR_IMPORT_KCODE(sys_close)
#endif
#if CR_KCODE_sys_ptrace
  #if !HAVE_SYS_PTRACE_PROTOTYPE
    /* Some kernels lack a prototype for sys_ptrace() */
    extern asmlinkage int sys_ptrace(unsigned long request, unsigned long pid, unsigned long addr, unsigned long data);
  #endif
  CR_IMPORT_KCODE(sys_ptrace)
#endif
#if defined(CR_KCODE_sys_setresuid) && CR_KCODE_sys_setresuid
  extern long sys_setresuid(uid_t, uid_t, uid_t);
  CR_IMPORT_KCODE(sys_setresuid)
#endif
#if defined(CR_KCODE_sys_setresgid) && CR_KCODE_sys_setresgid
  extern long sys_setresgid(gid_t, gid_t, gid_t);
  CR_IMPORT_KCODE(sys_setresgid)
#endif
#if defined(CR_KCODE_sys_setgroups) && CR_KCODE_sys_setgroups
  extern long sys_setgroups(int, gid_t *);
  CR_IMPORT_KCODE(sys_setgroups)
#endif
#if CR_KCODE_do_pipe
  CR_IMPORT_KCODE(do_pipe)
#endif
#if CR_KCODE_detach_pid
  CR_IMPORT_KCODE(detach_pid)
#endif
#if CR_KCODE_attach_pid
  CR_IMPORT_KCODE(attach_pid)
#endif
#if CR_KCODE_find_pid
  CR_IMPORT_KCODE(find_pid)
#endif
#if defined(CR_KCODE_link_pid) && CR_KCODE_link_pid
  CR_IMPORT_KCODE(link_pid)
#endif
#if defined(CR_KCODE___put_task_struct) && CR_KCODE___put_task_struct
  CR_IMPORT_KCODE(__put_task_struct)
#endif
#if defined(CR_KCODE___put_task_struct_cb) && CR_KCODE___put_task_struct_cb
  CR_IMPORT_KCODE(__put_task_struct_cb)
#endif
#if CR_KCODE_path_lookup
  CR_IMPORT_KCODE(path_lookup) // missing in early 2.4 kernels
#endif
#if CR_KCODE_vfs_read
  CR_IMPORT_KCODE(vfs_read)
#endif
#if CR_KCODE_vfs_write
  CR_IMPORT_KCODE(vfs_write)
#endif
#if CR_KCODE_copy_siginfo_to_user
  CR_IMPORT_KCODE(copy_siginfo_to_user)
#endif
#if CR_KCODE_set_fs_pwd
  CR_IMPORT_KCODE(set_fs_pwd)
#endif
#if CR_KCODE_put_filp
  CR_IMPORT_KCODE(put_filp)
#endif
#if CR_KCODE_sys_mprotect
  CR_IMPORT_KCODE(sys_mprotect)
#endif
#if CR_KCODE_do_exit
  CR_IMPORT_KCODE(do_exit)
#endif
#if CR_KCODE_do_sigaction
  CR_IMPORT_KCODE(do_sigaction)
#endif
#if CR_KCODE_do_execve
  CR_IMPORT_KCODE(do_execve)
#endif
#if defined(CR_KCODE_load_gs_index) && CR_KCODE_load_gs_index
  CR_IMPORT_KCODE(load_gs_index)  // x86_64
#endif
#if CR_KCODE_expand_files
  #if !HAVE_EXPAND_FILES_PROTOTYPE
    extern int expand_files(struct files_struct *, int);
  #endif
  CR_IMPORT_KCODE(expand_files)
#endif


MODULE_AUTHOR("Lawrence Berkeley National Lab " PACKAGE_BUGREPORT);
MODULE_DESCRIPTION("Berkeley Lab Checkpoint/Restart symbol imports kernel module");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

