/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_module.h,v 1.128 2006/06/13 20:36:17 phargrov Exp $
 */

#ifndef _CR_MODULE_H
#define _CR_MODULE_H        1

#include "blcr_config.h"	// Configuration

#include <linux/module.h>
#include <linux/config.h>

#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/version.h>
#include <linux/list.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/mount.h>
#include <linux/slab.h>
#include <linux/highmem.h>
#include <linux/spinlock.h>
#include <asm/atomic.h>
#if HAVE_LINUX_NAMEI_H
  #include <linux/namei.h>
#endif
#if HAVE_LINUX_SYSCALLS_H
  #include <linux/syscalls.h>           /* for sys_close, etc. */
#endif
#if HAVE_LINUX_SYSCALL_H
  #include <linux/syscall.h>           /* for sys_close, etc. */
#endif
#if HAVE_LINUX_RCUPDATE_H
  #include <linux/rcupdate.h>
#endif
#if defined(CONFIG_COMPAT)
  #include <linux/compat.h>
  /* This mapping works only because we only pass pointers and longs */
  #define CR_IOCTL32_MAP(x)	((((x)&0xff0000) == 0x80000) ? (((x)&~0xff0000)|0x40000) : (x))
  #define CR_IOCTL32_UNMAP(x)	((((x)&0xff0000) == 0x40000) ? (((x)&~0xff0000)|0x80000) : (x))
#endif

#include <blcr_common.h>	// For the stuff common to user space

#include "cr_arch.h"		// Architecture-specific bits
#include "cr_barrier.h"		// Code for kernel barriers.
#include "cr_ktrace.h"		// Code for trace messages


// cr_objectmap_t is an opaque type
struct cr_objectmap_s;
typedef struct cr_objectmap_s *cr_objectmap_t;

// Destination info for a checkpoint
typedef	struct cr_chkpt_dest_s {
	// file pointer (NULL iff destination is a directory)
	struct file *		filp;
	struct semaphore	mutex;

	// fs struct pointer (NULL iff destination is not a directory)
	struct fs_struct *	fs;
} cr_chkpt_dest_t;

typedef struct cr_rstrt_src_s {
        // file pointer (NULL iff destination is a directory)
        struct file *           filp;
        struct semaphore        mutex;

        // fs struct pointer (NULL iff destination is not a directory)
        struct fs_struct *      fs;
} cr_rstrt_src_t;

// MOD_{INC,DEC}_USE_COUNT are deprecated in 2.6.x
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0))
  #undef MOD_INC_USE_COUNT
  #define MOD_INC_USE_COUNT       do { } while (0)
  #undef MOD_DEC_USE_COUNT
  #define MOD_DEC_USE_COUNT       do { } while (0)
#endif

// Implement the 'struct work_struct' in terms of the older 'struct tq_struct' if needed
#if HAVE_WORK_STRUCT
  #if HAVE_LINUX_WORKQUEUE_H
    #include <linux/workqueue.h>
  #endif
  #define CR_INIT_WORK(_work, _task, _data, _func) \
	INIT_WORK((_work), (_task), (_data))
#elif HAVE_TQ_STRUCT
  struct work_struct {
      struct tq_struct	task;
      struct timer_list	timer;
  };
  #if HAVE_INIT_TQUEUE
    #ifndef INIT_TQUEUE
        #define INIT_TQUEUE(_tq, _func, _data) do {	\
	    INIT_LIST_HEAD(&(_tq)->list);		\
	    (_tq)->sync = 0;				\
	    (_tq)->routine = (_func);    		\
	    (_tq)->data = (_data);             		\
	  } while (0)
    #endif
    #define CR_INIT_WORK(_work, _task, _data, _func)  do {	\
	INIT_TQUEUE(&((_work)->task), (_task), (_data));	\
        init_timer(&((_work)->timer));				\
        (_work)->timer.function = (_func);			\
        (_work)->timer.data = (unsigned long)&(_work)->task;	\
      } while (0);
  #else
    #define CR_INIT_WORK(_work, _task, _data, _func)  do {	\
        INIT_LIST_HEAD(&((_work)->task.list));			\
        (_work)->task.sync = 0;					\
        (_work)->task.routine = &watchdog_task;			\
        (_work)->task.data = req;				\
        init_timer(&((_work)->timer));				\
        (_work)->timer.function = (_func);			\
        (_work)->timer.data = (unsigned long)&((_work)->task);	\
      } while (0);
  #endif
#else
  #error
#endif

#ifdef CAP_KILL
  #define cr_capable(X) capable(X)
#else
  #define cr_capable(X) suser()
#endif

#if HAVE_2_4_PERMISSION
  #define cr_permission(I,M,N)	permission((I),(M))
#elif HAVE_2_6_PERMISSION
  #define cr_permission(I,M,N)	permission((I),(M),(N))
#else
  #error
#endif

#if HAVE_TASK_RLIM
  #define CR_RLIM(task)	(task)->rlim
#elif HAVE_SIGNAL_RLIM
  #define CR_RLIM(task)	(task)->signal->rlim
#else
  #error
#endif

#ifndef rcu_assign_pointer
  #define rcu_assign_pointer(A,B)	((A) = (B))
#endif
#ifndef rcu_read_lock
  #define rcu_read_lock()		do {} while (0)
  #define rcu_read_unlock()		do {} while (0)
#endif

#if HAVE_STRUCT_FDTABLE
  typedef struct fdtable	cr_fdtable_t;
  #define cr_fdtable(files)	files_fdtable(files)
#else
  typedef struct files_struct	cr_fdtable_t;
  #define cr_fdtable(files)	(files)
#endif

// Kernel-side tracking of a checkpoint request
//
// XXX: this needs a lot of stuff added as we progress:
// It will also keep such things as the count of tasks in the
// checkpoint, reference counts for resources, etc.
//
// XXX: the naming between user and kernel could be improved.
typedef struct cr_chkpt_req_s {
	struct semaphore        serial_mutex;   // globally order steps in req
	atomic_t		ref_count;	// reference count
	atomic_t		completed;	// count of successful completions
	rwlock_t		lock;		// lock for task_list & wait
	wait_queue_head_t	wait;		// place for requester to wait
	struct list_head	tasks;		// list of target cr_task_t's
	cr_chkpt_dest_t		dest;		// checkpoint destination
	unsigned int		flags;		// flags supplied by requestor
	int			result;		// value returned by REAP
	int			has_expiration;	// 'expiration' is used
	unsigned long		expiration;	// limit on time to chkpt

	cr_objectmap_t          map;

	// barriers: XXX: could/should be done per thread group?
	cr_barrier_t		phase_barrier;		// between phases 1 & 2
	cr_barrier_t		predump_barrier;	// at start of dump
	cr_barrier_t		postdump_barrier;	// at end of dump

	struct work_struct	watchdog;	// Watchdog task and timer:

        cr_format_t             dump_format;    // Format to use for this dump

	cr_bool_t               done_header;
#if HAVE_2_6_SIGNAL_STRUCT
	cr_bool_t               done_signal_struct;
#endif
	cr_bool_t               done_linkage;
	cr_bool_t               done_creds;
	cr_bool_t               done_fs;
	cr_bool_t               done_files;
	cr_scope_t              checkpoint_scope;

	int			die;		// flag indicating we must die
	int			signal;		// sent after CHECKPOINT
} cr_chkpt_req_t;

typedef struct cr_rstrt_req_s {
        cr_rstrt_src_t          src;
        struct file            *file;
        cr_bool_t               done_fs;
        cr_bool_t               done_files;
	cr_bool_t               done_hide_cr_fds;
	cr_bool_t               done_close_cr_fds;
	cr_bool_t               done_creds;
	cr_bool_t               need_children;
	int	                clones_needed;
	int                     total_threads;
	int                     shared_signal;
	atomic_t                ref_count;
	struct task_struct     *cr_restart_task;
	struct cr_context_tasklinkage *linkage;
	int                     linkage_size;
	struct semaphore        serial_mutex;
	/* barriers for cr_restore_linkage */
	atomic_t		linkage_counter;
	cr_barrier_t		linkage_barrier;
	/* place for requester to wait for us to finish */
	wait_queue_head_t	wait;
	int			result;		// value returned by REAP
	int			die;		// indicates spawned threads must die
	cr_objectmap_t          map;
} cr_rstrt_req_t;

typedef enum {
	CR_NO_PHASE = 0,
	CR_PHASE1,
	CR_PHASE2
} cr_phase_t;

// For forming a list of tasks.
typedef struct cr_task_s {
	struct list_head	task_list;
	struct list_head	req_list;
	atomic_t		ref_count;
	struct task_struct	*task;
	cr_chkpt_req_t		*chkpt_req;
	cr_rstrt_req_t		*rstrt_req;
	cr_phase_t		phase;
	void			*handler;	// XXX: type?
	struct file		*filp;		// associated w/ handler (if any)
	int			stopped;
} cr_task_t;

// Private data attached to an instance of the file
typedef struct cr_pdata_s {
	cr_chkpt_req_t		*chkpt_req;
	// cr_chkpt_opt_t	*chkpt_opt;	NOT YET DEFINED
	cr_rstrt_req_t	        *rstrt_req;
} cr_pdata_t;

#if !HAVE_TIMEVAL_TO_JIFFIES
	// Basic adaptation of timespec_to_jiffies from include/linux/time.h
	static __inline__ unsigned long
	timeval_to_jiffies(struct timeval *value)
	{
		unsigned long sec = value->tv_sec;
		long usec = value->tv_usec;

		if (sec >= (MAX_JIFFY_OFFSET / HZ))
			return MAX_JIFFY_OFFSET;
		usec += 1000000L / HZ - 1;
		usec /= 1000000L / HZ;
		return HZ * sec + usec;
	}
#endif

#if !HAVE_JIFFIES_TO_TIMEVAL
	// Basic adaptation of jiffies_to_timeval from include/linux/time.h
	static __inline__ void
	jiffies_to_timeval(unsigned long jiffies, struct timeval *value)
	{
		value->tv_usec = (jiffies % HZ) * (1000000L / HZ);
		value->tv_sec = jiffies / HZ;
	}
#endif
	
#ifndef list_for_each_safe
// list_for_each_safe: A list iterator allowing for deletions
// First appears in 2.4.10 kernel, from which we've taken the following:
  #define list_for_each_safe(pos, n, head) \
    for (pos = (head)->next, n = pos->next;\
         pos != (head);\
         pos = n, n = pos->next)
#endif

// Task accessor macros (must be lvalues)
#if HAVE_2_6_TASK_IDS
  #define cr_task_pgrp(_t)	((_t)->signal->pgrp)
  #define cr_task_tty_old_pgrp(_t)	((_t)->signal->tty_old_pgrp)
  #define cr_task_session(_t)	((_t)->signal->session)
  #define cr_task_leader(_t)	((_t)->signal->leader)
#elif HAVE_2_4_TASK_IDS
  #define cr_task_pgrp(_t)	((_t)->pgrp)
  #define cr_task_tty_old_pgrp(_t)	((_t)->tty_old_pgrp)
  #define cr_task_session(_t)	((_t)->session)
  #define cr_task_leader(_t)	((_t)->leader)
#else
  #error
#endif

// Process table iterators
#if HAVE_DO_EACH_TASK_PID
  #define CR_DO_EACH_TASK_PGID(ID, T) \
    do_each_task_pid((ID), PIDTYPE_PGID, (T))
  #define CR_WHILE_EACH_TASK_PGID(ID, T) \
    while_each_task_pid((ID), PIDTYPE_PGID, (T))

  #define CR_DO_EACH_TASK_SID(ID, T) \
    do_each_task_pid((ID), PIDTYPE_SID, (T))
  #define CR_WHILE_EACH_TASK_SID(ID, T) \
    while_each_task_pid((ID), PIDTYPE_SID, (T))

  #define CR_DO_EACH_TASK_TGID(ID, T) \
    do_each_task_pid((ID), PIDTYPE_TGID, (T))
  #define CR_WHILE_EACH_TASK_TGID(ID, T) \
    while_each_task_pid((ID), PIDTYPE_TGID, (T))

  #define CR_DO_EACH_TASK_PROC(P, T) \
    if ((P)->mm) { \
      pid_t _id = cr_task_pgrp(P); \
      CR_DO_EACH_TASK_PGID(_id, (T)) \
        if ((T)->mm == (P)->mm)
  #define CR_WHILE_EACH_TASK_PROC(ID, T) \
      while_each_task_pid(_id, PIDTYPE_PGID, (T)); \
    }
#elif HAVE_FOR_EACH_TASK_PID
  #define CR_DO_EACH_TASK_PGID(ID, T) \
    do { \
      struct list_head *_l; \
      struct pid *_p; \
      for_each_task_pid((ID), PIDTYPE_PGID, (T), _l, _p)
  #define CR_WHILE_EACH_TASK_PGID(ID, T) \
    } while(0)

  #define CR_DO_EACH_TASK_SID(ID, T) \
    do { \
      struct list_head *_l; \
      struct pid *_p; \
      for_each_task_pid((ID), PIDTYPE_SID, (T), _l, _p)
  #define CR_WHILE_EACH_TASK_SID(ID, T) \
    } while(0)

  #define CR_DO_EACH_TASK_TGID(ID, T) \
    do { \
      struct list_head *_l; \
      struct pid *_p; \
      for_each_task_pid((ID), PIDTYPE_TGID, (T), _l, _p)
  #define CR_WHILE_EACH_TASK_TGID(ID, T) \
    } while(0)

  #define CR_DO_EACH_TASK_PROC(P, T) \
    if ((P)->mm) { \
      pid_t _id = cr_task_pgrp(P); \
      CR_DO_EACH_TASK_PGID(_id, (T)) { \
        if ((T)->mm == (P)->mm) {
  #define CR_WHILE_EACH_TASK_PROC(P, T) \
        } \
      } CR_WHILE_EACH_TASK_PGID(_id, (T)); \
    }

#elif HAVE_FOR_EACH_TASK
  #define CR_DO_EACH_TASK_PGID(ID, T) \
    for_each_task((T)) if (cr_task_pgrp(T) == (ID)) 
  #define CR_WHILE_EACH_TASK_PGID(ID, T)

  #define CR_DO_EACH_TASK_SID(ID, T) \
    for_each_task((T)) if (cr_task_session(T) == (ID)) 
  #define CR_WHILE_EACH_TASK_SID(ID, T)

  #define CR_DO_EACH_TASK_TGID(ID, T) \
    %%%ERROR%%% Use of TGID field with for_each_task
  #define CR_WHILE_EACH_TASK_TGID(ID, T) \
    %%%ERROR%%% Use of TGID field with for_each_task

  #define CR_DO_EACH_TASK_PROC(P, T) \
    if ((P)->mm) \
      for_each_task((T)) \
        if ((T)->mm == (P)->mm) 
  #define CR_WHILE_EACH_TASK_PROC(P, T)
#else
  #error No tasklist iterators available
#endif

/* What is the opposite of "get_task_struct" ? */
#if HAVE_FREE_TASK_STRUCT
  #define put_task_struct free_task_struct
#elif HAVE_PUT_TASK_STRUCT
  /* will CR_IMPORT_KSYM(__put_task_struct); */
#else
  #error
#endif

/* How do we manipulate the mmap_sem? */
#if HAVE_RW_MMAP_SEM
  #define cr_mmap_sem_down_read		down_read
  #define cr_mmap_sem_down_write	down_write
  #define cr_mmap_sem_up_read		up_read
  #define cr_mmap_sem_up_write		up_write
#else
  #define cr_mmap_sem_down_read		down
  #define cr_mmap_sem_down_write	down
  #define cr_mmap_sem_up_read		up
  #define cr_mmap_sem_up_write		up
#endif

/* How do we manipulate the file_lock */
#if HAVE_RW_FILE_LOCK
  #define cr_file_lock_read		read_lock
  #define cr_file_lock_unread		read_unlock
  #define cr_file_lock_write		write_lock
  #define cr_file_lock_unwrite		write_unlock
#elif HAVE_SPINLOCK_FILE_LOCK
  #define cr_file_lock_read		spin_lock
  #define cr_file_lock_unread		spin_unlock
  #define cr_file_lock_write		spin_lock
  #define cr_file_lock_unwrite		spin_unlock
#else
  #error "Unknown file_lock type"
#endif 

/* How do we manipulate the lock on a inode */
#if HAVE_INODE_SEM
  #define cr_inode_lock(_i)			down(&(_i)->i_sem)
  #define cr_inode_lock_interruptible(_i)	down_interruptible(&(_i)->i_sem)
  #define cr_inode_unlock(_i)			up(&(_i)->i_sem)
  #define cr_pipe_lock(_i)			down(PIPE_SEM(*(_i)))
  #define cr_pipe_lock_interruptible(_i)	down_interruptible(PIPE_SEM((*_i)))
  #define cr_pipe_unlock(_i)			up(PIPE_SEM(*(_i)))
#elif  HAVE_INODE_MUTEX
  #define cr_inode_lock(_i)			mutex_lock(&(_i)->i_mutex)
  #define cr_inode_lock_interruptible(_i)	mutex_lock_interruptible(&(_i)->i_mutex)
  #define cr_inode_unlock(_i)			mutex_unlock(&(_i)->i_mutex)
  #define cr_pipe_lock(_i)			mutex_lock(PIPE_MUTEX(*(_i)))
  #define cr_pipe_lock_interruptible(_i)	mutex_lock_interruptible(PIPE_MUTEX(*(_i)))
  #define cr_pipe_unlock(_i)			mutex_unlock(PIPE_MUTEX(*(_i)))
#else
  #error "Unknown inode lock type"
#endif 

#ifndef wait_event_interruptible_timeout
/* from 2.6.8 */
  #define __wait_event_interruptible_timeout(wq, condition, ret) \
    do {                                                         \
      DECLARE_WAITQUEUE(__wait, current);                        \
      add_wait_queue(&(wq), &__wait);                            \
      for (;;) {                                                 \
        set_current_state(TASK_INTERRUPTIBLE);                   \
        if (condition) break;                                    \
        if (!signal_pending(current)) {                          \
	  ret = schedule_timeout(ret);                           \
	  if (!ret) break;                                       \
	  continue;                                              \
        }                                                        \
        ret = -ERESTARTSYS;                                      \
        break;                                                   \
      }                                                          \
      current->state = TASK_RUNNING;                             \
      remove_wait_queue(&(wq), &__wait);                         \
    } while(0)
  #define wait_event_interruptible_timeout(wq, condition, timeout) \
    ({ long __ret = timeout;                                       \
       if (!(condition))                                           \
         __wait_event_interruptible_timeout(wq, condition, __ret); \
       __ret;                                                      \
    })
#endif

#if defined(EXIT_ZOMBIE)
  #define cri_task_dead(task)	 ((task)->exit_state & (EXIT_DEAD | EXIT_ZOMBIE))
#elif defined(TASK_DEAD)
  #define cri_task_dead(task)	 ((task)->state & (TASK_DEAD | TASK_ZOMBIE))
#else
  #define cri_task_dead(task)	 ((task)->state == TASK_ZOMBIE)
#endif

/* Prototypes for non-exported and/or non-prototyped kernel symbols */
extern void * sys_call_table[];
#if /* CR_KCODE_sys_dup2 && */ !HAVE_SYS_DUP2_PROTOTYPE
    /* Some kernels lack a prototype for sys_dup2() */
    extern asmlinkage long sys_dup2(unsigned int oldfd, unsigned int newfd);
#endif
#if /* CR_KCODE_sys_lseek && */ !HAVE_SYS_LSEEK_PROTOTYPE
    /* Some kernels lack a prototype for sys_lseek() */
    extern asmlinkage off_t sys_lseek(unsigned int fd, off_t offset, unsigned int origin);
#endif
#if /* CR_KCODE_sys_ftruncate && */ !HAVE_SYS_FTRUNCATE_PROTOTYPE
    /* Some kernels lack a prototype for sys_ftruncate() */
    extern asmlinkage long sys_ftruncate(unsigned int fd, unsigned long length);
#endif
#if CR_KDATA_write_pipe_fops
  extern struct file_operations write_pipe_fops;
#endif
#if CR_KDATA_rdwr_pipe_fops
  extern struct file_operations rdwr_pipe_fops;
#endif
#if CR_KDATA_read_pipe_fops
  extern struct file_operations read_pipe_fops;
#endif
#if CR_KDATA_write_fifo_fops
  extern struct file_operations write_fifo_fops;
#endif
#if CR_KDATA_rdwr_fifo_fops
  extern struct file_operations rdwr_fifo_fops;
#endif
#if CR_KDATA_read_fifo_fops
  extern struct file_operations read_fifo_fops;
#endif
#if CR_KDATA_anon_pipe_buf_ops
  extern struct pipe_buf_operations anon_pipe_buf_ops;
#endif
/* Next three may not be defined if not being used */
#if defined(CR_KCODE_sys_setresuid) && CR_KCODE_sys_setresuid
  extern long sys_setresuid(uid_t, uid_t, uid_t);
#endif
#if defined(CR_KCODE_sys_setresgid) && CR_KCODE_sys_setresgid
  extern long sys_setresgid(gid_t, gid_t, gid_t);
#endif
#if defined(CR_KCODE_sys_setgroups) && CR_KCODE_sys_setgroups
  extern long sys_setgroups(int, gid_t *);
#endif
#if /* CR_KCODE_expand_files && */ !HAVE_EXPAND_FILES_PROTOTYPE
  extern int expand_files(struct files_struct *, int);
#endif


// cr_proc.c
extern int cr_proc_init(void);
extern void cr_proc_cleanup(void);

// cr_fops.c
extern struct file_operations cr_ctrl_fops;
#if defined(CONFIG_COMPAT) && !defined(HAVE_COMPAT_IOCTL)
extern int cr_init_ioctl32(void);
extern int cr_cleanup_ioctl32(void);
#endif

// cr_dump_self.c
extern int cr_dump_self(struct file *filp, unsigned long flags);
extern int cr_cont_complete(struct file *filp);
extern char *cr_getpath(struct dentry *dentry, struct vfsmount *mnt,
			char *buf, int size);

// cr_chkpt_req.c
extern int cr_chkpt_req(struct file *filp, void *arg);
extern int cr_chkpt_done(struct file *filp, void *arg);
extern int cr_chkpt_reap(struct file *filp);
extern void cr_chkpt_task_complete(cr_task_t *cr_task);
extern void cr_chkpt_req_release(struct file *filp, cr_pdata_t *priv);
extern int cr_abort_checkpoint(struct file *filp, unsigned long flags);
extern int cr_chkpt_dest(struct file *filp, char *arg);

// cr_async.c
extern int cr_suspend(struct file *filp, void *arg);
extern int cr_phase1_register(struct file *filp, void *arg);
extern void cr_phase1_release(struct file *filp, cr_pdata_t *priv);

// cr_sync.c
extern int cr_phase2_register(struct file *filp, void *arg);
extern void cr_phase2_release(struct file *filp, cr_pdata_t *priv);

// cr_task.c
extern rwlock_t cr_task_lock;
extern struct list_head cr_task_list;
extern cr_task_t *__cr_task_get(struct task_struct *task, int create);
extern void __cr_task_put(cr_task_t *cr_task);
extern cr_task_t *cr_task_get(struct task_struct *task);
extern void cr_task_put(cr_task_t *cr_task);

// cr_rstrt_req.c
extern int cr_rstrt_request_restart(struct file *filp, void *arg);
extern int cr_rstrt_done(struct file *filp, void *arg);
extern int cr_rstrt_reap(struct file *filp);
extern int cr_rstrt_child(struct file *filp);
extern int cr_rstrt_clones(struct file *filp, unsigned long *new_stack_p);
extern int cr_rstrt_complete(struct file *filp);
extern int cr_rstrt_req_release(struct file *filp, cr_pdata_t *priv);

// cr_dest_file.c
extern int cr_dest_init(cr_chkpt_dest_t *dest, int fd, struct file *from);
extern void cr_dest_free(cr_chkpt_dest_t *dest);
extern struct file *cr_dest_get(cr_chkpt_dest_t *dest, int *shared);
extern void cr_dest_put(cr_chkpt_dest_t *dest, struct file *filp);

// cr_trigger.c
extern int cr_trigger_phase1(cr_chkpt_req_t *req);
extern int cr_trigger_phase2(cr_chkpt_req_t *req);

// cr_io.c
extern ssize_t cr_kread(struct file * file, void *buf, size_t count);
extern ssize_t cr_kwrite(struct file * file, void *buf, size_t count);

// cr_objects.c
extern cr_objectmap_t cr_alloc_objectmap(void);
extern void cr_release_objectmap(cr_objectmap_t);
extern int cr_find_object(cr_objectmap_t, void *, void **);
extern int cr_insert_object(cr_objectmap_t, void *, void *);
extern int cr_remove_object(cr_objectmap_t, void *);

#ifdef CONFIG_COMPAT
// cr_compat.c
extern long cr_compat_ctrl_ioctl(struct file *, unsigned, unsigned long);
#endif

#endif
